#include "syscall.h"

long data_value = 0xabadface;
int bss_value;

#ifdef USE_SYSCALL_WRAPPER
void __exit(int status)
{
  __r_exit(0, status);
}
int __read(int fd, void *buffer, int size)
{
  return __r_read(0, fd, buffer, size);
}
int __write(int fd, const void *buffer, int size)
{
  return __r_write(0, fd, buffer, size);
}
int __open(const char *path, int flags, int mode)
{
  return __r_open(0, path, flags, mode);
}
int __close(int fd)
{
  return __r_close(0, fd);
}
#endif

/* System call functions */

void _exit(int status)
{
  __exit(status);
}

int read(int fd, void *buffer, int size)
{
  int i, r;

  for (i = 0; i < size; i += r) {
    r = __read(fd, (char *)buffer + i, size - i);
    if (r < 0)
      return r;
    if (r == 0)
      break;
  }

  return i;
}

int write(int fd, const void *buffer, int size)
{
  int i, r;

  for (i = 0; i < size; i += r) {
    r = __write(fd, (char *)buffer + i, size - i);
    if (r < 0)
      return r;
    if (r == 0)
      break;
  }

  return i;
}

int open(const char *path, int flags, int mode)
{
#ifdef OPENFLAGS
  flags = OPENFLAGS;
#endif
  return __open(path, flags, mode);
}

int close(int fd)
{
  return __close(fd);
}

/* Library functions */

void exit(int status)
{
  _exit(status);
}

int getchar(int fd)
{
  unsigned char ch;
  return (read(fd, &ch, 1) > 0) ? ch : -1;
}

int putchar(int fd, int c)
{
  unsigned char ch = c;
  return (write(fd, &ch, 1) > 0) ? ch : -1;
}

int gets(int fd, char *str)
{
  int c, i = 0;

  while (1) {
    c = getchar(fd);
    if (c < 0)
      break;
    if (c == '\r') continue; /* 0x0d */
    str[i++] = c;
    if (c == '\n') break; /* 0x0a */
  }
  str[i] = '\0';

  return i;
}

int puts(int fd, char *str)
{
  int i;
  for (i = 0; str[i]; i++)
    ;
  return write(fd, str, i);
}

#ifdef RSHIFT
/*
 * For architecture not to be able to use shift instruction for well,
 * execute shift operation without shift instruction.
 */
unsigned long rshift1(unsigned long value)
{
  unsigned long mask, bit = 1;
  unsigned long ret = 0;
  int i;

  for (i = 0; i < sizeof(unsigned long) * 8 - 1; i++) {
    mask = bit + bit; /* Use addition for no using shift and multiplication */
    if (value & mask)
      ret |= bit;
    bit = bit + bit;
  }

  return ret;
}

unsigned long RSHIFT(unsigned long value)
{
  int num = 4;
  for (; num > 0; num--)
    value = rshift1(value);
  return value;
}
#endif

int putxval(int fd, unsigned long value, int column)
{
  char buf[17];
  char *p;

  p = buf + sizeof(buf) - 1;
  *(p--) = '\0';

  if (!value && !column)
    column++;

  while (value || column) {
    *(p--) = "0123456789abcdef"[value & 0xf];
    value >>= 4;
    if (column) column--;
  }

  puts(fd, p + 1);

  return 0;
}

int main()
{
  int fd = 1;
  puts(fd, "Hello World! ");
  putxval(fd, data_value, 0);
  puts(fd, " This architecture is " ARCH "\n");
  exit(0);
  return 0;
}
